package cascade

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Cascade.CoverSource

Suffix coverage table for same-selector shadowing.

Sourcetype t

Properties written later, keyed by selector.

Sourceval v : unit -> t

v () creates an empty coverage table.

Sourceval covered : t -> Selector.t -> Declaration.t -> bool

covered t selector decl is true when decl's property is definitely shadowed by a later declaration for selector at the same or stronger importance.

Sourceval add : t -> Selector.t -> Declaration.t -> unit

add t selector decl records that decl's property is written later for selector.